// Copyright 2021 Contributors to the Parsec project.
// SPDX-License-Identifier: Apache-2.0
mod algorithm_property_list_tests;
mod command_code_attributes_list_tests;
mod command_code_list_tests;
mod digest_list_tests;
mod ecc_curve_list_tests;
mod handle_list_tests;
mod pcr_selection_list_builder_tests;
mod pcr_selection_list_tests;
mod tagged_pcr_property_list_tests;
mod tagged_tpm_property_list_tests;
