// Copyright 2021 Contributors to the Parsec project.
// SPDX-License-Identifier: Apache-2.0
mod asymmetric_primitives;
mod attached_components;
mod attestation_commands;
mod authenticated_countdown_timer;
mod capability_commands;
mod clocks_and_timers;
mod command_audit;
mod context_management;
mod dictionary_attack_functions;
mod duplication_commands;
mod enhanced_authorization_ea_commands;
mod ephemeral_ec_keys;
mod field_upgrade;
mod hash_hmac_event_sequences;
mod hierarchy_commands;
mod integrity_collection_pcr;
mod miscellaneous_management_functions;
mod non_volatile_storage;
mod object_commands;
mod random_number_generator;
mod session_commands;
mod signing_and_signature_verification;
mod startup;
mod symmetric_primitives;
mod testing;
mod vendor_specific;
