// Copyright 2021 Contributors to the Parsec project.
// SPDX-License-Identifier: Apache-2.0
mod auth_handle_tests;
mod object_handle_tests;
mod pcr_handle_tests;
mod session_handle_tests;
mod tpm_handles_tests;
